import math
from collections import defaultdict

import numpy as onp

from avici.synthetic import Data
from avici.utils.graph import mat_to_toporder


def draw_rff_params(*, rng, d, length_scale, output_scale, n_rff):
    """Draws random instantiation of rffs"""
    # draw parameters
    ls = length_scale(rng, shape=(1,)).item() if callable(length_scale) else length_scale
    c = output_scale(rng, shape=(1,)).item() if callable(output_scale) else output_scale

    # draw rffs
    # [d, n_rff]
    omega_j = rng.normal(loc=0, scale=1.0 / ls, size=(d, n_rff))

    # [n_rff, ]
    b_j = rng.uniform(0, 2 * onp.pi, size=(n_rff,))

    # [n_rff, ]
    w_j = rng.normal(loc=0, scale=1.0, size=(n_rff,))

    return dict(
        c=c,
        omega=omega_j,
        b=b_j,
        w=w_j,
        n_rff=n_rff,
    )


def sample_recursive_scm(*,
                         rng,
                         n_observations_obs,
                         n_observations_int,
                         g,
                         f,
                         nse,
                         interv_dist,
                         n_interv_vars=0):
    """Ancestral sampling over a DAG

    Args:
        rng:
        n_observations_obs: number of observational data rows to be sampled
        n_observations_int: number of interventional data rows to be sampled
        g: adjacency matrix of the DAG of shape [n_vars, n_vars]
        f: list of functions (mechanisms), one for each node.
            Each f[j] maps: observation matrix `x` [n_obs, n_vars], noise vector `z` [n_obs,], and
            parent indicator `is_parent` [n_vars,] to the observations observed for
            node j [n_obs,]
        nse: list of class instances representing the noise distributions for each node, subclassing `NoiseModel` ABC
        interv_dist: Subclass of `DistributionModel` ABC for sampling intervention values
        n_interv_vars (optional): number of variables intervened upon (default is 0). If -1 is passed, all variables are intervened
            upon. For other integers, a set of intervened variables is randomly selected and interventional data
            is generated in equal proportion for each node based on the total number of `n_observations_int` data points

    Returns:
        dict containing `g`, `x_obs`, `x_int`, `n_vars`, `n_observations_obs`, `n_observations_int`, `is_count_data`
    """
    n_vars = g.shape[-1]
    toporder = mat_to_toporder(g)

    # sample target nodes for the interventions
    interv_targets = []

    simulate_observ_data = n_observations_obs > 0
    if simulate_observ_data:
        interv_targets += [None]

    simulate_interv_data = n_observations_int > 0
    if simulate_interv_data:
        assert n_interv_vars != 0, f"Need n_interv_vars != 0 to sample interventional data"
        if n_interv_vars == -1 or n_interv_vars == 1.0:
            n_interv_vars = n_vars
        elif not n_interv_vars.is_integer():
            n_interv_vars = math.ceil(n_interv_vars * n_vars)
        interv_targets += sorted(rng.choice(n_vars, size=min(n_vars, n_interv_vars), replace=False).tolist())

    assert (n_interv_vars == -1) or (0 <= n_interv_vars <= n_vars),\
        f"Got `n_interv_vars` = {n_interv_vars} for `n_vars` = {n_vars}, which is invalid."

    # simulate data for observational data and for each interventional target
    data = defaultdict(lambda: defaultdict(list))
    for interv_target in interv_targets:

        if interv_target is None:
            # observational
            data_type = "obs"
            is_intervened = onp.zeros(n_vars).astype(bool)
            n_obs = n_observations_obs

        else:
            # interventional
            data_type = "int"
            is_intervened = onp.eye(n_vars)[interv_target].astype(bool)
            n_obs = math.ceil(n_observations_int / n_interv_vars)

        # ancestral sampling in topological order
        x = onp.zeros((n_obs, n_vars))
        for j in toporder:
            # sample noise
            z_j = nse[j](rng=rng, x=x, is_parent=g[:, j])

            # compute node given parents and noise or perform intervention state
            if is_intervened[j]:
                x[:, j] = interv_dist(rng, shape=z_j.shape)
            else:
                x[:, j] = f[j](x=x, z=z_j, is_parent=g[:, j])

        # generate intervention mask
        # [n_obs, n_vars] with True/False depending on whether node was intervened upon
        interv_mask = onp.tile(is_intervened, (x.shape[0], 1)).astype(onp.float32)

        data[data_type]["x"].append(x)
        data[data_type]["interv_mask"].append(interv_mask)


    # concatenate interventional data, interweaving rows to have balanced observation counts when clipping the end
    if simulate_observ_data:
        x_obs = onp.stack(data["obs"]["x"]).reshape(-1, n_vars, order="F")
        x_obs_msk = onp.stack(data["obs"]["interv_mask"]).reshape(-1, n_vars, order="F")
    else:
        x_obs = onp.zeros((0, n_vars))  # dummy
        x_obs_msk = onp.zeros((0, n_vars))  # dummy

    if simulate_interv_data:
        x_int = onp.stack(data["int"]["x"]).reshape(-1, n_vars, order="F")
        x_int_msk = onp.stack(data["int"]["interv_mask"]).reshape(-1, n_vars, order="F")
    else:
        x_int = onp.zeros((0, n_vars))  # dummy
        x_int_msk = onp.zeros((0, n_vars))  # dummy

    # clip number of observations to have invariant shape (in case n_obs doesn't evenly devide no. interv targets)
    # [n_observations, n_vars, 2]
    x_obs = onp.stack([x_obs, x_obs_msk], axis=-1)[:n_observations_obs, :, :]
    x_int = onp.stack([x_int, x_int_msk], axis=-1)[:n_observations_int, :, :]

    assert x_obs.size != 0 or x_int.size != 0, f"Need to sample at least some observations; " \
                                               f"got shapes x_obs {x_obs.shape} x_int {x_int.shape}"

    # collect data
    data = Data(
        x_obs=x_obs,
        x_int=x_int,
        is_count_data=False,
    )
    return data


import warnings
warnings.formatwarning = lambda msg, category, path, lineno, file: f"{path}:{lineno}: {category.__name__}: {msg}\n"

import jax.numpy as jnp
import jax.random as random
import jax.lax as lax

"""Standardization"""

def _jax_cpm_standardizer(x):
    # compute library sizes (sum of row)
    x_libsize = x.sum(-1, keepdims=True)

    # divide each cell by library size and multiply by 10^6
    # will yield nan for rows with zero expression and for zero expression entries
    log2cpm = jnp.where(jnp.isclose(x, 0.0), jnp.nan, jnp.log2(x / (x_libsize * 1e-6)))
    return log2cpm


def _jax_cpm_shift_scale(x, shift, scale):
    # shift and scale
    x = (x - jnp.where(jnp.isnan(shift), 0.0, shift)) / jnp.where(jnp.isnan(scale), 1.0, scale)

    # set nans (set for all-zero rows, i.e. zero libsize) to minimum (i.e. to zero)
    # catch empty arrays (when x_int is empty and has axis n=0)
    if not x.size == 0:
        x = jnp.where(jnp.isnan(x), 0.0, x)

    return x


def _jax_standardize_count(x_obs, x_int):
    # cpm normalization
    x_obs = x_obs.at[..., 0].set(_jax_cpm_standardizer(x_obs[..., 0]))
    x_int = x_int.at[..., 0].set(_jax_cpm_standardizer(x_int[..., 0]))

    # subtract min (~robust global median) and divide by global std dev
    global_ref_x = jnp.concatenate([x_obs[..., 0], x_int[..., 0]], axis=-2)
    global_min = jnp.nanmin(global_ref_x, axis=(-1, -2), keepdims=True)
    global_std = jnp.nanstd(global_ref_x, axis=(-1, -2), keepdims=True)

    x_obs = x_obs.at[..., 0].set(_jax_cpm_shift_scale(x_obs[..., 0], global_min, global_std))
    x_int = x_int.at[..., 0].set(_jax_cpm_shift_scale(x_int[..., 0], global_min, global_std))

    return x_obs, x_int


def jax_standardize_count_simple(x):
    """_jax_standardize_count but with only one argument"""
    # cpm normalization
    x = x.at[..., 0].set(_jax_cpm_standardizer(x[..., 0]))

    # subtract min (~robust global median) and divide by global std dev
    global_ref_x = x[..., 0]
    global_min = jnp.nanmin(global_ref_x, axis=(-1, -2), keepdims=True)
    global_std = jnp.nanstd(global_ref_x, axis=(-1, -2), keepdims=True)

    x = x.at[..., 0].set(_jax_cpm_shift_scale(x[..., 0], global_min, global_std))
    return x


def _jax_standardize_default(x_obs, x_int):
    # default z-standardization
    ref_x = jnp.concatenate([x_obs[..., 0], x_int[..., 0]], axis=-2)
    mean = ref_x.mean(-2, keepdims=True)
    std = ref_x.std(-2, keepdims=True)
    x_obs_default = x_obs.at[..., 0].set((x_obs[..., 0] - mean) / jnp.where(std == 0.0, 1.0, std))
    x_int_default = x_int.at[..., 0].set((x_int[..., 0] - mean) / jnp.where(std == 0.0, 1.0, std))
    return x_obs_default, x_int_default


def jax_standardize_default_simple(x):
    """_jax_standardize_default but with only one argument"""
    # default z-standardization
    ref_x = x[..., 0]
    mean = ref_x.mean(-2, keepdims=True)
    std = ref_x.std(-2, keepdims=True)
    x = x.at[..., 0].set((x[..., 0] - mean) / jnp.where(std == 0.0, 1.0, std))
    return x


def jax_standardize_data(data):
    """Standardize observations `x_obs` and `x_int`"""
    x_obs = data["x_obs"]
    x_int = data["x_int"]

    assert (x_obs.shape[-1] == 1 or x_obs.shape[-1] == 2) \
       and (x_int.shape[-1] == 1 or x_int.shape[-1] == 2), \
        f"Assume concat 3D shape but got: x_obs {x_obs.shape} and x_int {x_int.shape}"

    x_obs_count, x_int_count = _jax_standardize_count(x_obs, x_int)
    x_obs_default, x_int_default = _jax_standardize_default(x_obs, x_int)
    return {
        **data,
        "x_obs": jnp.where(data["is_count_data"][..., None, None, None], x_obs_count, x_obs_default),
        "x_int": jnp.where(data["is_count_data"][..., None, None, None], x_int_count, x_int_default),
    }

def jax_standardize_x(x, is_count_data):
    """Standardize observations"""
    assert (x.shape[-1] == 1 or x.shape[-1] == 2),\
        f"Assume concat 3D shape but got: x {x.shape}"

    # TODO this is not a nice implementation; refactor jax_standardize_data to accept variable number of x fields in data
    dummy = jnp.zeros((*x.shape[:-3], 0, *x.shape[-2:]))
    data = jax_standardize_data({"x_obs": x, "x_int": dummy, "is_count_data": is_count_data})
    return data["x_obs"]


"""Data batching"""

def jax_get_train_x(key, batch, p_obs_only):
    key, subk = random.split(key)
    n_obs_and_int, n_int = batch["x_obs"].shape[-3], batch["x_int"].shape[-3]

    only_observational = random.bernoulli(key, p_obs_only)
    x = lax.cond(
        only_observational,
        # only observational data
        lambda _: batch["x_obs"], # already has N=n_obs + n_int
        # mix observational and interventional
        # select `n_obs` elements of `n_obs + n_int` available observational data
        lambda _: jnp.concatenate([
            batch["x_obs"][..., random.permutation(subk, n_obs_and_int)[:(n_obs_and_int - n_int)], :, :],
            batch["x_int"]
        ], axis=-3),
        operand=None,
    )

    key, subk = random.split(key)
    x = x[..., random.permutation(key, n_obs_and_int), :, :]
    x = jax_standardize_x(x, batch["is_count_data"])  # place 1/2 where jax data should be standardized; do not do anywhere else to avoid double-standardizing count data
    return x


def jax_get_x(batch):
    x = jnp.concatenate([batch["x_obs"], batch["x_int"]], axis=-3)
    x = jax_standardize_x(x, batch["is_count_data"]) # place 2/2 where jax data should be standardized; do not do anywhere else to avoid double-standardizing count data
    return x


import inspect
from abc import ABCMeta, abstractmethod
from typing import NamedTuple, Any


class PrintMixin:
    def __repr__(self):
        cls = self.__class__.__name__
        idhex = hex(id(self))
        attrs = " ".join(f"{k}={v!r}" for k, v in self.__dict__.items())
        attrs = ": " + attrs if attrs else ""
        return f"<{cls} at {idhex}{attrs}>"


class AbstractSignatureChecker(ABCMeta):
    """
    Meta class for strictly enforcig signatures of @abstractmethod's in an abstract base class
    https://stackoverflow.com/a/55315285
    """
    def __init__(cls, name, bases, attrs):
        errors = []
        for base_cls in bases:
            for meth_name in getattr(base_cls, "__abstractmethods__", ()):
                orig_argspec = inspect.getfullargspec(getattr(base_cls, meth_name))
                target_argspec = inspect.getfullargspec(getattr(cls, meth_name))
                if orig_argspec != target_argspec:
                    errors.append(
                        f"Subclass `{cls.__name__}` of `{base_cls.__name__}` not implemented with correct signature "
                        f"in abstract method {meth_name!r}.\n"
                        f"Expected: {orig_argspec}\n"
                        f"Got: {target_argspec}\n")
        if errors:
            raise TypeError("\n".join(errors))
        super().__init__(name, bases, attrs)


class Distribution(PrintMixin, metaclass=AbstractSignatureChecker):
    """
    Abstract base class for a distribution
    """
    @abstractmethod
    def __call__(self, rng, shape=None):
        """

        Args:
            rng (np.random.Generator): numpy pseudorandom number generator
            shape (tuple, optional): tuple shape for sampling

        Returns:
            ndarray: sample of shape `shape` and `()` if `None`
        """
        pass


class GraphModel(PrintMixin, metaclass=AbstractSignatureChecker):
    """
    Abstract base class for graph model p(G)
    """
    @abstractmethod
    def __call__(self, rng, n_vars):
        """

        Args:
            rng (np.random.Generator): numpy pseudorandom number generator
            n_vars (int): number of nodes in the graph

        Returns:
            ndarray: binary adjacency matrix of shape `[n_vars, n_vars]`
        """
        pass


class MechanismModel(PrintMixin, metaclass=AbstractSignatureChecker):
    """
    Abstract base class for data-generating mechanism p(D|G)
    """
    @abstractmethod
    def __call__(self, rng, g, n_observations_obs, n_observations_int):
        """

        Args:
            rng (np.random.Generator): numpy pseudorandom number generator
            g (ndarray): binary adjacency matrix of shape `[n_vars, n_vars]` as generated by a `GraphModel` subclass
            n_observations_obs (int): number of observational data points to be sampled
            n_observations_int (int): number of interventional data points to be sampled

        Returns:
            avici.synthetic.Data: namedtuple containing `x_obs` and `x_int` data matrices of
                `n_observations_obs` and `n_observations_int` rows, respectively, as well as a boolean `is_count_data`
                indicating whether the data are counts.

                The data matrices `x_obs` and `x_int` should have shapes `[n_observations_obs, n_vars, 2]` and
                `[n_observations_int, n_vars, 2]`, respectively. The first value in the last axis
                (i.e. `x_int[..., 0]`) contains the values and the second axis (i.e. `x_int[..., 1]`) contains
                either 0 or 1, indicating which nodes were intervened upon in which observations.
                Accordingly, `x_obs[..., 1]` has only zeros as it always contains observational data.
                If `n_observations_obs` or `n_observations_int` are zero, a numpy array should still be retuyrned
                with size 0 in the respective first axis (allowing both to later be concatened).

                The boolean `is_count_data` is used to determine how the data is standardized.
                (Default for all real-valued data should be `False`, which implies default z-standardization.)

        """
        pass


class NoiseModel(PrintMixin, metaclass=AbstractSignatureChecker):
    """
    Abstract base class for SCM noise
    """
    @abstractmethod
    def __call__(self, rng, x, is_parent):
        """

        Args:
            rng (np.random.Generator): numpy pseudorandom number generator
            x (ndarray): data matrix of shape `[n_observations, n_vars]`
            is_parent (ndarray): binary vector of shape `[n_vars,]` indicating which nodes are parents of this node

        Returns:
            ndarray: noise sampled of shape `[n_observations]`

        """
        pass


class CustomClassWrapper(NamedTuple):
    name: str
    kwargs: Any
    paths: Any


class SyntheticSpec(NamedTuple):
    """
    Data structure for a single component of the data-generating distribution.
    Elements of the YAML domain configuration `data` should match this signature.

    Args:
        graph (GraphModel):
        mechanism (MechanismModel):
        n_observations_obs (int):
        n_observations_int (int):

    """
    graph: Any
    mechanism: Any
    n_observations_obs: int = None
    n_observations_int: int = 0


class Data(NamedTuple):
    """
    Data structure for returning data from a `MechanismModel`

    Args:
        x_obs (ndarray): observational data matrix of shape `[n_observations_obs, n_vars, 2]`.
            The first value in the last axis (i.e. `x_obs[..., 0]`) contains the values and
            the second axis (i.e. `x_obs[..., 1]`) contains either 0 or 1, indicating whether
            the node was intervened upon in this observation. Accordingly, since `x_obs` contains
            only observational data, `x_obs[..., 0]` always has only zeros.
        x_int (ndarray): interventional data matrix of shape `[n_observations_int, n_vars, 2]`
            The first value in the last axis (i.e. `x_int[..., 0]`) contains the values and
            the second axis (i.e. `x_int[..., 1]`) contains either 0 or 1, indicating whether
            the node was intervened upon in this observation.
        is_count_data (bool): indicates whether the data are counts, which should probably only be `True`
            for gene expression data (`GRNSergio`). This flag is used to determine how the data is standardized.
            (Default for all real-valued data should be `False`, which implies default z-standardization.)

    """
    x_obs: Any
    x_int: Any
    is_count_data: bool





import subprocess
import sys
import argparse
from datetime import datetime


def get_version():
    git_commit = subprocess.check_output(["git", "describe", "--always"]).strip().decode(sys.stdout.encoding) 
    return "_" + git_commit


def get_datetime():
    now = datetime.now()
    date_time = now.strftime("%Y-%m-%d-%H%M")
    return "_" + date_time


def str2bool(v):
    v = "".join([char for char in v if u"\xa0" not in char])  # avoid utf8 parsing error
    if isinstance(v, bool):
       return v
    if v.lower() in ('yes', 'True', 'true', 'T', 't', 'Y', 'y', '1'):
        return True
    elif v.lower() in ('no', 'False', 'false', 'F', 'f', 'N', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')



import time
import math
import numpy as onp
import pyarrow.plasma as plasma

from typing import NamedTuple, Any
from avici.utils.data import onp_subbatch
from avici.utils.parse import init_custom_classes


class SharedArray(NamedTuple):
    max_size: Any
    state: Any
    mem: Any
    state_mem_lock: Any
    metrics: Any
    metrics_lock: Any


class SharedState(NamedTuple):
    queues: Any
    buffers: Any
    is_alive: Any
    next_job: Any
    plasma_store_name: Any


class Sampler:

    dtype = dict(
        g=onp.int32,
        n_vars=onp.int32,
        x_obs=onp.float32,
        x_int=onp.float32,
        n_observations_obs=onp.int32,
        n_observations_int=onp.int32,
        is_count_data=bool,
    )

    @staticmethod
    def generate_data(rng, *, n_vars, spec_list, spec=None):
        # sample random spec
        if spec is None:
            spec = spec_list[rng.choice(len(spec_list))]

        # sample graph sampling
        g = spec.graph(rng=rng, n_vars=n_vars)

        # sample mechanism mechanism and data
        data = spec.mechanism(rng=rng,
                              g=g,
                              n_observations_obs=spec.n_observations_obs,
                              n_observations_int=spec.n_observations_int)

        # collect all data
        collection = dict(
            g=g,
            n_vars=n_vars,
            x_obs=data.x_obs,
            x_int=data.x_int,
            n_observations_obs=spec.n_observations_obs,
            n_observations_int=spec.n_observations_int,
            is_count_data=data.is_count_data,
        )

        # check dims
        if collection["x_obs"].ndim == 2:
            collection["x_obs"] = onp.expand_dims(collection["x_obs"], axis=-1)
        if collection["x_int"].ndim == 2:
            collection["x_int"] = onp.expand_dims(collection["x_int"], axis=-1)

        return collection

    @staticmethod
    def async_worker(j, state, worker_seed, it, descr, n_vars, config):
        """Runs `f` repeatedly and sends results to the q.
        Arguments to worker must not be functions or generators.
        To see errors for debugging, must call job.get() """

        rng = onp.random.default_rng(worker_seed)

        # initialize custom classes inside handle pickling with multiprocessing and custom modules
        config = init_custom_classes(config)

        # init plasma client
        plasma_client = plasma.connect(state.plasma_store_name)

        ctr = j
        for _ in it or infinite_iterator():
            ctr += 1
            if not state.is_alive.is_set():
                break

            # get buffer for which to generate data
            assert n_vars is not None or descr == "train"
            if n_vars is None:
                d = state.next_job.value
            else:
                d = n_vars

            with state.buffers[descr][d].metrics_lock:
                state.buffers[descr][d].metrics.items_added += 1

            # generate data
            data = Sampler.generate_data(
                rng,
                n_vars=d,
                spec_list=config["data"][descr],
            )

            # commit data to plasma store
            object_id = plasma_client.put(data)

            # push object id on random processing queue
            # avoiding querying load status
            q = state.queues[ctr % len(state.queues)]
            q.put(dict(descr=descr, n_vars=d, id=object_id))

        # disconnect plasma client
        plasma_client.disconnect()


class FIFOBuffer(Sampler):

    @staticmethod
    def init(manager, buffer_size):
        max_size = buffer_size

        state = manager.Namespace()
        state.index = 0
        state.size = 0

        metrics = manager.Namespace()
        metrics.items_added = 0
        metrics.items_sampled = 0

        mem = manager.list([None] * buffer_size)

        state_mem_lock = manager.Lock()
        metrics_lock = manager.Lock()

        return SharedArray(
            max_size=max_size,
            state=state,
            mem=mem,
            state_mem_lock=state_mem_lock,
            metrics=metrics,
            metrics_lock=metrics_lock,
        )

    @staticmethod
    def load_balanc_async(state, t_sleep=0.001):
        # small worker that continually updates buffer with highest sample-insert ratio
        while True and state.is_alive.is_set():
            si = {d: get_soft_sample_insert_ratio(buffer) for d, buffer in state.buffers["train"].items()}
            state.next_job.value = max(si, key=si.get)
            time.sleep(t_sleep)

    @staticmethod
    def insert_async(j, state):

        # init plasma client
        plasma_client = plasma.connect(state.plasma_store_name)
        q = state.queues[j]

        while True:
            # get next item from q
            item = q.get()
            if item == 'kill':
                q.task_done()
                break  # kill signal for listener

            b = state.buffers[item["descr"]][item["n_vars"]]

            # FIFO buffer update
            with b.state_mem_lock:
                old_object_id = b.mem[b.state.index]
                b.mem[b.state.index] = item["id"]

                b.state.size = min(b.state.size + 1, b.max_size)
                b.state.index = (b.state.index + 1) % b.max_size

            # remove old data from plasma store
            plasma_client.delete([old_object_id])

            q.task_done()

        # disconnect plasma client
        plasma_client.disconnect()

    @staticmethod
    def index_into_buffer(j, *, client, buffer):
        with buffer.state_mem_lock:
            return client.get(buffer.mem[j])

    @staticmethod
    def uniform_sample_from_buffer(_, *, rng, client, buffer, n_obs, n_int):
        with buffer.state_mem_lock:
            idx = int(rng.choice(buffer.state.size))
            item_id = buffer.mem[idx]
            assert client.contains(item_id)
            item = client.get(item_id, timeout_ms=100)
            assert item != plasma.ObjectNotAvailable

        with buffer.metrics_lock:
            buffer.metrics.items_sampled += 1

        subitem = onp_subbatch(rng, item, n_obs=n_obs, n_int=n_int)
        return subitem


def infinite_iterator():
    """An infinite iterator (enabling for loop like `while True`)"""
    while True:
        yield None


def get_soft_sample_insert_ratio(buffer, const=1e3):
    # softness const ensures smooth load balancing initially
    if buffer.metrics.items_added:
        return (buffer.metrics.items_sampled + const) / buffer.metrics.items_added
    else:
        return math.inf


def is_buffer_filled(buffer):
    return all([item is not None for item in buffer.mem])


import numpy as onp
from avici.synthetic.utils import draw_rff_params
from avici.synthetic import NoiseModel


class SimpleNoise(NoiseModel):
    """
    Simple noise with constant scale (homogeneous noise)

    Args:
        dist (Distribution): distribution from which noise is sampled
        scale (float): scale by which noise is scaled (similar to standard deviation in Gaussian)

    """
    def __init__(self, dist, scale):
        self.dist = dist
        self.scale = scale

    def __call__(self, rng, x, is_parent):
        n_obs = x.shape[0]
        return self.scale * self.dist(rng, shape=(n_obs,))


class HeteroscedasticRFFNoise(NoiseModel):
    """
    Heterogenous noise based on a random Fourier feature scaling function h(.)
    The noise scale of `dist` is input-dependent an given by scale^2 = log(1 + exp(h(x))

    Args:
        dist (Distribution): distribution from which noise is sampled
        rng (np.random.Generator): numpy pseudorandom number generator
        d (int): input dimension (number of parents)
        length_scale (Distribution): distribution from which length scale of GP is sampled
        output_scale (Distribution): distribution from which length scale of GP is sampled
        n_rff (int, optional): number of features used to approximate GP function draw
    """

    def __init__(self, dist, *, rng, d, length_scale, output_scale, n_rff=100):
        self.dist = dist
        self.param = draw_rff_params(rng=rng, d=d, length_scale=length_scale,
                                     output_scale=output_scale, n_rff=n_rff)
        self.n_rff = n_rff

    def __call__(self, rng, x, is_parent):
        # compute rff function f(x) = w'phi(x)
        x_parents = x[:, onp.where(is_parent)[0]]
        phi = onp.cos(onp.einsum('db,...d->...b', self.param["omega"], x_parents) + self.param["b"])
        f_x = onp.sqrt(2.0) * self.param["c"] * onp.einsum('b,...b->...', self.param["w"], phi) / onp.sqrt(self.n_rff)

        # sample noise with scale^2 = log(1+exp(f(x))
        scale = onp.sqrt(onp.log(1.0 + onp.exp(f_x)))
        return scale * self.dist(rng, shape=scale.shape)


def init_noise_dist(*, rng, dim, dist, noise_scale, noise_scale_heteroscedastic):

    if noise_scale is not None:
        assert noise_scale_heteroscedastic is None

        scale = noise_scale(rng)
        return SimpleNoise(dist, scale)

    elif noise_scale_heteroscedastic is not None:
        assert noise_scale is None
        assert "rff" in noise_scale_heteroscedastic

        return HeteroscedasticRFFNoise(dist, rng=rng, d=int(dim),
                                       length_scale=noise_scale_heteroscedastic["length_scale"],
                                       output_scale=noise_scale_heteroscedastic["output_scale"],
                                       n_rff=100)
    else:
        raise KeyError("neither `noise_scale` nor `noise_scale_heteroscedastic` are given")


from setuptools import setup, find_packages
setup(
    name='avici',
    version='1.0.4',
    description='Amortized Inference for Causal Structure Learning',
    author='Lars Lorch',
    author_email='lars.lorch@inf.ethz.ch',
    url="https://github.com/larslorch/avici",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    package_data={
        'avici': [
            'assets/*.tsv',
            'config/examples/*.yaml',
            'config/train/*.yaml',
            'synthetic/sergio/*.yaml',
        ],
    },
    install_requires=[
        'jax>=0.3.17',
        'jaxlib>=0.3.14',
        'imageio',
        'jupyter',
        'matplotlib',
        'numpy',
        'scipy',
        'pandas',
        'igraph',
        'scikit-learn',
        'sklearn',
        'tqdm>=4.64.1',
        'dm-haiku>=0.0.8',
        'psutil',
        'optax>=0.1.3',
        'pyarrow==7.0.0',
        'tensorflow-datasets~=4.3.0',
        'deepdiff',
        'seaborn',
        'torch',
        'huggingface-hub',
    ]
)



import sys
import avici
sys.modules['amortibs'] = avici # legacy compatibility with old package name with pickled checkpoints
sys.modules['amortibs.modules'] = avici # legacy compatibility with old package name with pickled checkpoints

import warnings
from pathlib import Path
import yaml
import json
import pickle

from avici.definitions import CHECKPOINT_KWARGS, PROJECT_DIR


def load_yaml(path, abspath=False):
    """Load plain yaml config"""

    load_path = path if abspath else (PROJECT_DIR / path)

    try:
        with open(load_path, "r") as stream:
            try:
                config = yaml.safe_load(stream)
                return config

            except yaml.YAMLError as exc:
                warnings.warn(f"YAML parsing error. Returning `None` for config.\n")
                print(exc, flush=True)
                return None
    except FileNotFoundError:
        warnings.warn(f"{Path(path).name} doesn't exist. Returning `None` for config. "
                      f"This is fine if it is the data.yaml in a train experiment folder.\n")
        return None


def load_checkpoint(folder):
    # load state
    last_checkpoint = sorted([p for p in folder.iterdir() if ".pkl" in p.name])[-1]
    with open(last_checkpoint, 'rb') as file:
        state = pickle.load(file)

    # load kwargs
    kwargs_checkpoint = folder / CHECKPOINT_KWARGS
    with open(kwargs_checkpoint, "r") as file:
        kwargs = json.load(file)
        kwargs["inference_model_kwargs"]["acyclicity"] = None

    return state, kwargs
